"======================================================================
|
|   Float Method Definitions
|
|
 ======================================================================"


"======================================================================
|
| Copyright 1988,92,94,95,99,2000,2001,2002
| Free Software Foundation, Inc.
| Written by Steve Byrne.
|
| This file is part of the GNU Smalltalk class library.
|
| The GNU Smalltalk class library is free software; you can redistribute it
| and/or modify it under the terms of the GNU Lesser General Public License
| as published by the Free Software Foundation; either version 2.1, or (at
| your option) any later version.
| 
| The GNU Smalltalk class library is distributed in the hope that it will be
| useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
| General Public License for more details.
| 
| You should have received a copy of the GNU Lesser General Public License
| along with the GNU Smalltalk class library; see the file COPYING.LIB.
| If not, write to the Free Software Foundation, 59 Temple Place - Suite
| 330, Boston, MA 02111-1307, USA.  
|
 ======================================================================"


Number variableByteSubclass: #Float
       instanceVariableNames: ''
       classVariableNames: ''
       poolDictionaries: 'CSymbols'
       category: 'Language-Data types'
!

Float comment: 
'My instances represent floating point numbers that have 64 bits of 
precision (well, less than that in precision; they are precisely the same
as C''s "double" datatype).  Besides the standard numerical operations,
I provide transcendental operations too.'  !


!Float class methodsFor: 'converting'!

coerce: aNumber
    "Answer aNumber converted to a Float"
    ^aNumber asFloat
! !


!Float class methodsFor: 'byte-order dependancies'!

signByte
    "Answer the byte of the receiver that contains the exponent"
    ^Bigendian ifTrue: [ 1 ] ifFalse: [ CDoubleSize ]
! !


!Float class methodsFor: 'characterization'!

radix
    "Answer the base in which computations between instances of the receiver 
     are made.  This should be 2 on about every known computer, so GNU
     Smalltalk always answers 2."
    ^2
!

denormalized
    "Answer whether instances of the receiver can be in denormalized
     form."
    ^self smallestDenormalized > 0
!

e
    "Returns the value of e. Hope is that it is precise enough"
    ^2.7182818284590452353602874713527d0
!

precision
    "Answer the number of bits in the mantissa. 1 + (2^-precision) = 1"
    ^CDoubleBinaryDigits
!

fminNormalized
    "Return the smallest normalized Float that is > 0"
    ^CDoubleMin
!

fminDenormalized
    "Return the smallest Float that is > 0 if denormalized values
     are supported, else return 0."
    ^self fminNormalized timesTwoPower: 1 - self precision
!

fmin
    "Return the smallest Float that is > 0."
    | fminDen fmin |
    fmin := self fminNormalized.
    fminDen := fmin timesTwoPower: 1 - self precision.
    ^fminDen = 0 ifTrue: [ fmin ] ifFalse: [ fminDen ].
!

fmax
    "Return the largest normalized Float that is not infinite."
    ^CDoubleMax
!

emax
    "Return the maximum allowable exponent for a Float that is finite."
    ^CDoubleMaxExp
!

emin
    "Return the maximum allowable exponent for a Float that is finite."
    ^CDoubleMinExp
!

decimalDigits
    "Return the number of decimal digits of precision for a Float.
     Technically, if P is the precision for the representation, then
     the decimal precision Q is the maximum number of decimal digits
     such that any floating point number with Q base 10 digits can be
     rounded to a floating point number with P base 2 digits and back
     again, without change to the Q decimal digits."
    ^CDoubleDigits
!

epsilon
    "Return the smallest Float x for which is 1 + x ~= 1"
    ^(self coerce: 2.0) timesTwoPower: self precision negated
!

log10Base2
    "Returns the value of log2 10. Hope is that it is precise enough"
    ^3.3219280948873623478703194294894d0
!

ln10
    "Returns the value of ln 10. Hope is that it is precise enough"
    ^2.3025850929940456840179914546844d0
!

infinity
    "Return a Float that represents positive infinity."
    ^CDoublePInf
!

negativeInfinity
    "Return a Float that represents negative infinity."

    ^CDoubleNInf
!

nan
    "Return a Float that represents a mathematically indeterminate value
     (e.g. Inf - Inf, Inf / Inf)."
    ^CDoubleNaN
!

pi
    "Returns the value of pi. Hope is that it is precise enough"
    ^3.14159265358979323846264338327950288d0
! !



!Float methodsFor: 'arithmetic'!

negated
    "Return the negation of the receiver.  Unlike 0-self, this converts
     correctly signed zeros."
    ^self * -1
!

integerPart
    "Return the receiver's integer part"
    ^self - self fractionPart
! !



!Float methodsFor: 'testing'!

isNaN
    "Answer whether the receiver represents a NaN"
    ^self ~= self
!

isFinite
    "Answer whether the receiver does not represent infinity, nor a NaN"
    ^(self * self zero) = (self * self zero)
!

isInfinite
    "Answer whether the receiver represents positive or negative infinity"
    ^self = Float infinity or: [ self = Float negativeInfinity ]
!

negative
    "Answer whether the receiver is negative"
    ^(self at: self class signByte) > 127
!

strictlyPositive
    "Answer whether the receiver is > 0"
    ^(self at: self class signByte) < 128 and: [ self ~= self zero ]
!

positive
    "Answer whether the receiver is positive.  Negative zero is
     not positive, so the definition is not simply >= 0."
    ^(self at: self class signByte) < 128
!

sign
    "Answer 1 if the receiver is greater than 0, -1 if less than 0,
     else 0.  Negative zero is the same as positive zero."

    self = self zero ifTrue: [^0].
    ^self < 0 ifTrue: [-1] ifFalse: [1]
! !



!Float methodsFor: 'coercing'!

zero
    "Coerce 0 to the receiver's class"
    ^0.0
!

unity
    "Coerce 1 to the receiver's class"
    ^1.0
!

coerce: aNumber
    "Coerce aNumber to the receiver's class"
    ^aNumber asFloat
!

generality
    "Answer the receiver's generality"
    ^400
!

asExactFraction
    "Convert the receiver into a fraction with optimal approximation,
     but with usually huge terms."

    self checkCoercion.

    ^(self   timesTwoPower:  self exponent + Float precision) truncated /
     (1      bitShift:       self exponent + Float precision)
!

asFraction
    "Convert the receiver into a fraction with a good (but undefined)
     approximation"

    | a x n2 d2 n1 d1 n0 d0 eps abs |

    self checkCoercion.

    "This uses an algorithm based on the theory of continued fractions!!!
      n2/d2 = numerator and denominator of the fraction two steps ago
      n1/d1 = numerator and denominator of the fraction a steps ago
      n0/d0 = numerator and denominator of the fraction at the current step"

    n1  := d0 := 0.
    n0  := d1 := 1.
    abs := self abs.
    eps := Float epsilon * 1024.	"Number out of a hat"
    x   := abs.

    [   a := x truncated.
	n2 := n1. d2 := d1.
	n1 := n0. d1 := d0.
	n0 := n1 * a + n2.
	d0 := d1 * a + d2.
	(n0 asFloat / d0 asFloat - abs) abs < eps
    ]   whileFalse: [
	x := x fractionPart reciprocal
    ].

    ^Fraction
	numerator: (self < 0 ifTrue: [ n0 negated ] ifFalse: [ n0 ])
	denominator: d0
!


estimatedLog
    "Answer an estimate of (self abs floorLog: 10)"
    ^(self exponent + 1) asFloat / Float log10Base2
!

asFloat
    "Just defined for completeness.  Return the receiver."
    ^self
! !


!Float methodsFor: 'printing'!

printOn: aStream
    "Print a representation of the receiver on aStream"
    self
	printOn: aStream
	special: #('Inf' '-Inf' 'NaN')
! !



!Float methodsFor: 'storing'!

storeOn: aStream
    "Print a representation of the receiver on aStream"
    self
	printOn: aStream
	special: #('%1 infinity' '%1 negativeInfinity' '%1 nan')
! !



!Float methodsFor: 'private'!

checkCoercion
    "Private - Fail if the receiver is only representable as a Float"

    self isInfinite
	ifTrue: [self arithmeticError: 'Infinity can only be a Float'].
    self isNaN
	ifTrue: [self arithmeticError: 'Not-a-Number can only be a Float'].
!

exponentLetter
    "Private - Return the letter to be printed just before the exponent"
    ^$e
!

printOn: aStream special: whatToPrintArray
    "Private - Print a decimal representation of the receiver on aStream,
     printing one of the three elements of whatToPrintArray if it is
     infinity, negative infinity, or a NaN"
    | num str exp digits decimalDigits noExpSmallest noExpLargest |

    self isNaN ifTrue: [
	^aStream nextPutAll:
	    ((whatToPrintArray at: 3) bindWith: aStream class) ].
    self = self class infinity ifTrue: [
	^aStream nextPutAll:
	    ((whatToPrintArray at: 1) bindWith: aStream class) ].
    self = self class negativeInfinity ifTrue: [
	^aStream nextPutAll:
	    ((whatToPrintArray at: 2) bindWith: aStream class) ].

    self negative ifTrue: [ aStream nextPut: $- ].
    self = 0.0q0 ifTrue: [ aStream nextPutAll: '0.0'. ^self ].

    num := self abs.

    decimalDigits := 10 raisedToInteger: self class decimalDigits - 1.
    noExpSmallest := 10.0q0 raisedToInteger: 1 - self class decimalDigits.
    noExpLargest := 10.0q0 raisedToInteger: self class decimalDigits.

    (num >= noExpSmallest and: [ num < noExpLargest ]) ifFalse: [
	"Print in scientific notation"
	exp := 0.
	num < 1.0q0
	    ifTrue: [
		[ num := num * 10.0q0. exp := exp - 1. num < 1.0q0 ] whileTrue
	    ]
	    ifFalse: [
		[ num := num / 10.0q0. exp := exp + 1. num < 10.0q0 ] whileFalse
 	    ].

	"Produce the digits, up to a maximum of 14 = -1 + log10 (2^52)."
	num := (num * decimalDigits) asInteger.
	num = decimalDigits
	    ifTrue: [ num := num // 10. exp := exp + 1 ].

	str := num printString.
	str := str copyFrom: 1 to: (2 max: (str findLast: [ :each | each ~= $0 ])).

	str keysAndValuesDo: [ :n :ch |
	    aStream nextPut: ch.
	    n = 1 ifTrue: [ aStream nextPut: $. ].
	].
	aStream nextPut: self exponentLetter; print: exp.
	^self
    ].

    "Print a not-so-big number"
    num >= 1.0q0 ifTrue: [
	str := num truncated printString.
	aStream nextPutAll: str; nextPut: $. .

	digits := self class decimalDigits - str size.
	digits = 0 ifTrue: [ ^self ].
	num := num fractionPart *
	    (10 raisedToInteger: digits).

	str := num truncated printString.
	digits := digits - str size.
	str := str copyFrom: 1 to: (1 max: (str findLast: [ :each | each ~= $0 ])).
	aStream next: digits put: $0; nextPutAll: str.
	^self
    ].

    "Print a not-so-small number"
    str := (num * decimalDigits) rounded printString.
    aStream nextPutAll: '0.'.

    digits := self class decimalDigits - 1 - str size.
    str := str copyFrom: 1 to: (1 max: (str findLast: [ :each | each ~= $0 ])).
    aStream next: digits put: $0; nextPutAll: str.
! !


!Float methodsFor: 'testing functionality'!

isFloat
    ^true
! !

Smalltalk at: #FloatE put: Float!
Smalltalk at: #FloatD put: Float!
Smalltalk at: #FloatQ put: Float!

